#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#define BUFLEN 512
#define NPACK 10
#define PORT 9930
//#define SRV_IP "127.0.01"
#define SRV_IP "192.168.1.40"

#define DEBUG 0


#include "TicTacToe.h"
#include "TicTacToeMsgs.h"

#define bool int
#define true 1
#define false 0

#define SERVER 0x01
#define CLIENT 0x02

#ifdef DEBUG
FILE * logFile;
#endif
int counter = 0;
int playType = 0;
bool done = 0;
struct Point tokenLocations[] = {
                           {2,1},{2,6}, {2,11},
                           {7,1},{7,6}, {7,11},
                           {12,1},{12,6}, {12,11},
                         };

int xLoc = 0;
int yLoc = 3;

unsigned int xSelections = 0;
unsigned int oSelections = 0;

int turn = 0;
int gameOver = 0;
/*************************************************************/
void diep(char *s)
{
  perror(s);
  exit(1);
}

/*************************************************************/
int BuildIt(char _a, char _b)
{
  int y = 0;
  y = (_b<<8) | (_a & 0x00ff);
//  printf("Decoded y: %d\n", y);
  return y;
}

/*************************************************************/
void ConvertInt(int x, char *_a, char *_b)
{
  *_a = x & 0xff;
  *_b = (x>>8) ;
}


/*************************************************************/

struct sockaddr_in si_me, si_other;
int playerSocket; 

void InitServerSocket();
void InitClientSocket();
void InitScreen();
void ProcessInputs();
int PlaceToken(int _loc);
void EvalBoard();
void BuildMessage(char * _buf);
void ProcessArgs(int argc, char ** argv);
void RedrawBoard();
void ProcessMsg(char * buf);

/*************************************************************/
int main(int argc, char ** argv)
{
  ProcessArgs(argc, argv);
  if (playType == SERVER)
  {
#ifdef DEBUG
    logFile = fopen("server.log", "w");
#endif
    InitServerSocket();
    printf ("PlayType : SERVER\n");
    turn = XTURN;
  }
  else if (playType == CLIENT)
  {
#ifdef DEBUG
    logFile = fopen("client.log", "w");
#endif
    turn = OTURN;
    InitClientSocket();
    printf ("PlayType : CLIENT\n");
  }
  else
  {
    printf("PlayType : UNKNOWN\n");
    exit(0);
  }
  system("clear");
  while (!done)
  {
    gameOver = false;
    xLoc = 0;
    yLoc = 3;
    if (playType == SERVER)
    {
      turn = XTURN;
    }
    else if (playType == CLIENT)
    {
      turn = OTURN;
    }
//    turn = 0;
    xSelections = 0;
    oSelections = 0;

    InitScreen();
    while (!gameOver)
    {
      if (playType == CLIENT)
      {
        WaitForOpponent();
        EvalBoard();
        if (gameOver)
          break;
        ProcessInputs();
      }
      else
      {
        ProcessInputs();
        EvalBoard();
        if (gameOver)
          break;
        WaitForOpponent();
        EvalBoard();
      }
    }  
  }
//  printf("\x1b[4;10H");
//  printf("Hello!\n");
  return 0;
}
/*************************************************************/
void WaitForOpponent()
{
#ifdef DEBUG
  fprintf(logFile, "WaitForOpp ent\n");
  fflush(logFile);
#endif
  int s, i, j, slen=sizeof(si_other);
  char buf[BUFLEN];

  printf("\x1b[20;10H");
  printf("Wait for opp!\n");
  if (recvfrom(playerSocket, buf, BUFLEN, 0, &si_other, &slen)==-1)
    diep("recvfrom()");
//  printf("Received packet from %s:%d\nData: %s\n\n",
//    inet_ntoa(si_other.sin_addr), ntohs(si_other.sin_port), buf);
  printf("\x1b[20;10H");
  printf("            !\n");
  ProcessMsg(buf);
  if (buf[0] == PLAYERMOVE)
  {
    //ack...
    RedrawBoard();
    printf("\x1b[19;30H");
    printf("Received MOVE cmd");
  }
  else
  {
    printf("\x1b[19;30H");
    printf("Received Unknown cmd");
  }
#ifdef DEBUG
  fprintf(logFile, "WaitForOpp exit\n");
  fflush(logFile);
#endif
}
/*************************************************************/
void ProcessMsg(char * buf)
{
  if (buf[0] == PLAYERMOVE)
  {
    if (buf[5] == ENDOFMESSAGE)
    {
      xSelections = BuildIt(buf[1], buf[2]);
      oSelections = BuildIt(buf[3], buf[4]);
    }
  }
}

/*************************************************************/
void InitServerSocket()
{
  int s, i, slen=sizeof(si_other);
  char buf[BUFLEN];

  if ((playerSocket=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1)
    diep("socket");

  memset((char *) &si_me, sizeof(si_me), 0);
  si_me.sin_family = AF_INET;
  si_me.sin_port = htons(PORT);
  si_me.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(playerSocket, &si_me, sizeof(si_me))==-1)
    diep("bind");
  printf("SOCKET UP\n");
  if (recvfrom(playerSocket, buf, BUFLEN, 0, &si_other, &slen)==-1)
    diep("recvfrom()");
  printf("Client COnnected!\n");
  
}
/*************************************************************/
void InitClientSocket()
{

  int s, i, j, slen=sizeof(si_other);
  char buf[BUFLEN];

  if ((playerSocket=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1)
   diep("socket");

  memset((char *) &si_other, sizeof(si_other), 0);
  si_other.sin_family = AF_INET;
  si_other.sin_port = htons(PORT);
  if (inet_aton(SRV_IP, &si_other.sin_addr)==0)
  {
    fprintf(stderr, "inet_aton() failed\n");
    exit(1);
  }
  printf("Sending packet %d\n", i);
  sprintf(buf, "This is packet %d\n", i);
  if (sendto(playerSocket, buf, BUFLEN, 0, &si_other, slen)==-1)
    diep("sendto()");

}

/*************************************************************/
void SendMove()
{
#ifdef DEBUG
  fprintf(logFile, "Send Move ent\n");
  fflush(logFile);
#endif
  int slen = sizeof(si_other);
  char buf[512];
  memset(buf, 0, 512);
  BuildMessage(buf);
//  printf("Received packet from %s:%d\nData: %s\n\n",
//    inet_ntoa(si_other.sin_addr), ntohs(si_other.sin_port), buf);
  if (sendto(playerSocket,buf,BUFLEN,0,&si_other,slen) == -1)
    printf("Could not send packet!\n");
  printf("\x1b[19;10H");
  printf("Message Sent      ");
//  if (recvfrom(playerSocket, buf, BUFLEN, 0, &si_other, &slen)==-1)
//    diep("recvfrom()");
#ifdef DEBUG
  fprintf(logFile, "Send Move Exit\n");
  fflush(logFile);
#endif
}

/*************************************************************/
void BuildMessage(char * _buf)
{
  _buf[0] = PLAYERMOVE;
  ConvertInt(xSelections, &_buf[1], &_buf[2]);
  ConvertInt(oSelections, &_buf[3], &_buf[4]);
  _buf[5] = ENDOFMESSAGE; 
}

/*************************************************************/
void ProcessInputs()
{
#ifdef DEBUG
  fprintf(logFile, "ProcessInputs Ent\n");
  fflush(logFile);
#endif
  printf("\x1b[19;10H");
  printf("                  \n");
  printf("\x1b[20;10H");
  printf("ProcessInputs!\n");
//  printf("\x1b[20;20H");
  bool success = false;
  while (!success)
  {
    char ch = getchar();
    int temp = atoi(&ch);
    if (temp > 0)
      success = PlaceToken(temp-1);
  }
#ifdef DEBUG
  fprintf(logFile, "ProcessInputs Exit\n");
  fflush(logFile);
#endif
//  printf("\x1b[20;0H");
//  printf("CH:%x\n", ch); 
}
void RedrawBoard()
{
  int x;
  for (x = 0; x<9; x++)
  {
    if (xSelections & 1<<x)
    {
      printf("\x1b[%d;%dH", tokenLocations[x].y, tokenLocations[x].x);
      printf("%c", turnChars[0]);//magic number?
    }
    else if (oSelections & 1<<x)
    {
      printf("\x1b[%d;%dH", tokenLocations[x].y, tokenLocations[x].x);
      printf("%c", turnChars[1]);//magic number?
    }
  } 
}
/*************************************************************/
int PlaceToken(int _loc)
{
  printf("\x1b[0;0H");
  printf("%d", _loc);
  int spot = _loc;
  if (spot > -1)
  {
    if (!(xSelections & 1<<spot) && !(oSelections & 1<<spot))
    {
      printf("\x1b[%d;%dH", tokenLocations[spot].y, tokenLocations[spot].x);
//      pspDebugScreenSetXY(tokenLocations[spot].x, tokenLocations[spot].y);
      printf("%c", turnChars[turn]);
      if (turn == 0)
      {
        xSelections |= 1<<spot;
//        turn = OTURN;
      }
      else
      {
        oSelections |= 1<<spot;
//        turn = XTURN;
      }
      SendMove();
      EvalBoard();
      return true;
    }
  }
  return false;
}
/*************************************************************/
void ProcessArgs(int argc, char ** argv)
{
  int c;
  while((c = getopt(argc, argv, "c:s:")) != EOF){
    switch(c) {
    case 's':
      playType = SERVER;
      break;
    case 'c':
      playType = CLIENT;
      break;
    }
  }
}

/*************************************************************/

void EvalBoard()
{
//  pspDebugScreenSetXY(30,30);
  printf("\x1b[30;30H");
  int boardState = CheckForWinner(xSelections,oSelections);
  if (boardState == -1)
    return;
  if (boardState == 2)
  {
	printf("Draw!");
	gameOver = true;

    return;
  }
  if (boardState == 0)
  {
    printf("XWINS!");
	gameOver = true;
	//XWins!

    return;
  }
  if (boardState == 1)
  {
    printf("OWINS!");
	gameOver = true;
	//yWins!
    return;
  }

}

/*************************************************************/
void InitScreen()
{
  printf("\x1b[0;0H");
//  pspDebugScreenSetXY(0,0);
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("---------------\n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("---------------\n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("    |    |     \n");
  printf("    |    |     \n");

}
/*************************************************************/
